package it.sephiroth.ui.components
{
	import flash.display.BitmapData;
	import flash.display.BlendMode;
	import flash.display.DisplayObject;
	import flash.display.Shape;
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.events.MouseEvent;
	import flash.geom.Matrix;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	
	import flashx.textLayout.events.UpdateCompleteEvent;

	public class DistortTool extends Sprite
	{
		private var _controls_sprite: Sprite;
		private var _display_shape: Shape;
		private var _overlay: Shape;
		private var _down_control: DistortControlPoint;
		private var _index: Vector.<int>;
		private var _mouse_down: Boolean;
		private var _ne: DistortControlPoint;
		private var _no: DistortControlPoint;
		private var _points: Vector.<Vector.<Point>>;
		private var _se: DistortControlPoint;
		private var _segments: int = 20;
		private var _so: DistortControlPoint;
		private var _source: BitmapData;
		private var _target: DisplayObject;
		private var _uvdata: Vector.<Number>;
		private var _vertex: Vector.<Number>;
		private var _displayPoints: Boolean;

		public function DistortTool( nsegments: int = 20 )
		{
			super();
			_segments = nsegments;
			_display_shape = new Shape();
			_controls_sprite = new Sprite();
			_overlay = new Shape();
			
			addChild( _display_shape );
			addChild( _overlay );
			addChild( _controls_sprite );
			
			_overlay.visible = false;
			_overlay.blendMode = BlendMode.OVERLAY;
						
			addEventListener( Event.ADDED_TO_STAGE, onAdded );
			addEventListener( Event.REMOVED_FROM_STAGE, onRemoved );
		}

		public function get displayPoints():Boolean
		{
			return _displayPoints;
		}

		public function set displayPoints(value:Boolean):void
		{
			_displayPoints = value;
			_overlay.visible = value;
			
			if( _target )
			{
				_setDistort( noPoint, nePoint, soPoint, sePoint );
			}
		}
		
		protected function get noPoint(): Point
		{
			return _points[0][0];
		}
		
		protected function get nePoint(): Point
		{
			return _points[_segments][0];
		}
		
		protected function get soPoint(): Point
		{
			return _points[0][_segments];
		}
		
		protected function get sePoint(): Point
		{
			return _points[_segments][_segments];
		}
		
		public function get outputDisplay(): Shape
		{
			return _display_shape;
		}
		
		public function get outputBounds(): Rectangle
		{
			return _display_shape.getBounds( _display_shape );
		}

		public function get target(): DisplayObject
		{
			return _target;
		}

		public function get segments(): int
		{
			return _segments;
		}
		
		public function set segments( value: int ): void
		{
			if( _segments != value )
			{
				
				if( _target )
				{
					var no: Point = noPoint.clone();
					var ne: Point = nePoint.clone();
					var so: Point = soPoint.clone();
					var se: Point = sePoint.clone();
					_segments = value;
					
					updateData( true );
					setDistort( no, ne, so, se );
					setDistort( no, ne, so, se );
				}
				
				_segments = value;
			}
		}
		
		public function setDistort( no: Point, ne: Point, so: Point, se: Point ): void
		{
			_no.x = no.x;
			_no.y = no.y;
			_ne.x = ne.x;
			_ne.y = ne.y;
			_so.x = so.x;
			_so.y = so.y;
			_se.x = se.x;
			_se.y = _se.y;
			_setDistort( no, ne, so, se );
		}

		public function set target( value: DisplayObject ): void
		{
			setNewTarget( value );
		}

		private function initializeTarget(): void
		{
			if ( _source )
			{
				_source.dispose();
				_source = null;
			}

			while ( _controls_sprite.numChildren )
			{
				_controls_sprite.getChildAt( 0 ).removeEventListener( MouseEvent.MOUSE_DOWN, onMouseDown );
				_controls_sprite.removeChildAt( 0 );
			}
			
			_display_shape.graphics.clear();
			_se = _so = _no = _ne = null;

			if ( _target != null )
			{
				_source = new BitmapData( _target.width, _target.height, true, 0 );
				_source.draw( _target );

				updateData( false );
			}
		}
		
		private function updateData( update: Boolean ): void
		{
			createData( update );
			
			_no = _controls_sprite.getChildAt( 0 ) as DistortControlPoint;
			_ne = _controls_sprite.getChildAt( 2 ) as DistortControlPoint;
			_so = _controls_sprite.getChildAt( 1 ) as DistortControlPoint;
			_se = _controls_sprite.getChildAt( 3 ) as DistortControlPoint;
			
			_setDistort( noPoint, nePoint, soPoint, sePoint );
		}
		
		private function createData( update: Boolean = false ): void
		{
			_points = new Vector.<Vector.<Point>>();
			_vertex = new Vector.<Number>();
			_index = new Vector.<int>();
			_uvdata = new Vector.<Number>();
			
			var px: Number;
			var k: int = 0;
			var uvx: Number;
			var uvy: Number;
			
			for ( var x: int = 0; x < _segments + 1; ++x )
			{
				_points[x] = new Vector.<Point>();
				px = ( x / _segments ) * _target.width;
				uvx = x / _segments;
				
				for ( var y: int = 0; y < _segments + 1; y++ )
				{
					var py: Number = ( y / _segments ) * _target.height;
					_points[x][y] = new Point( px, py );
					uvy = y / _segments;
					
					if ( y < _segments && x < _segments )
					{
						_index.push( k, k + _segments + 1, k + 1, k + _segments + 1, k + _segments + 2, k + 1 );
					}
					
					_uvdata.push( uvx, uvy );
					
					if (( x == 0 || x == _segments ) && ( y == 0 || y == _segments ))
					{
						if( !update )
						{
							var c: DistortControlPoint = new DistortControlPoint();
							c.x = _points[x][y].x;
							c.y = _points[x][y].y;
							c.blendMode = BlendMode.INVERT;
							c.addEventListener( MouseEvent.MOUSE_DOWN, onMouseDown );
							_controls_sprite.addChild( c );
						}
					}
					++k;
				}
			}
		}

		private function onAdded( event: Event ): void
		{
		}

		protected function _setDistort( no: Point, ne: Point, so: Point, se: Point ): void
		{
			if ( _target )
			{
				if( _displayPoints )
				{
					_overlay.graphics.clear();
					_overlay.graphics.lineStyle( 0, 0xFFFFFF );
					_overlay.graphics.moveTo( 0, 0 );
				}
				
				_vertex.length = 0;
				var dist1: Number = ne.y - no.y;
				var dist2: Number = se.y - so.y;
				var dist3: Number = so.x - no.x;
				var dist4: Number = se.x - ne.x;
				var ystep_min: Number = ( so.y - no.y ) / _segments;
				var xstep_min: Number = ( ne.x - no.x ) / _segments;
				var ystep1: Number = 0;
				var ystep2: Number = 0;
				var xstep1: Number = 0;
				var xstep2: Number = 0;
				var pt: Point;
				var k: int = 0;

				for ( var x: int = 0; x < _segments+1; ++x )
				{
					xstep1 = 0;
					xstep2 = 0;

					for ( var y: int = 0; y < _segments+1; ++y )
					{
						pt = _points[x][y];
						pt.y = no.y + ystep1 + ( ystep_min * y ) + ((( ystep2 / _segments ) * y ) - (( ystep1 / _segments ) * y ));
						pt.x = no.x + xstep1 + ( xstep_min * x ) + ((( xstep2 / _segments ) * x ) - (( xstep1 / _segments ) * x ));
						
						if( _displayPoints )
						{
							_overlay.graphics.lineTo( pt.x, pt.y );
							
							if( x < _segments )
								_overlay.graphics.lineTo( _points[x+1][y].x, _points[x+1][y].y );
						}
						
						_vertex.push( pt.x, pt.y );
						xstep1 += ( dist3 / _segments );
						xstep2 += ( dist4 / _segments );
						
						k += 2;
						
					}
					ystep2 += ( dist2 / _segments );
					ystep1 += ( dist1 / _segments );
				}
				
				if( _displayPoints )
					_overlay.graphics.endFill();
				
				_display_shape.graphics.clear();
				_display_shape.graphics.beginBitmapFill( _source, null, false, true );
				_display_shape.graphics.drawTriangles( _vertex, _index, _uvdata );
			}
		}

		private function onMouseDown( event: MouseEvent ): void
		{
			stage.addEventListener( MouseEvent.MOUSE_MOVE, onMouseMove );
			stage.addEventListener( MouseEvent.MOUSE_UP, onMouseUp );
			_mouse_down = true;
			_down_control = DistortControlPoint( event.target );
		}

		private function onMouseMove( event: MouseEvent ): void
		{
			if ( _mouse_down )
			{
				_down_control.x = _controls_sprite.mouseX;
				_down_control.y = _controls_sprite.mouseY;
				
				_setDistort( _no.point, _ne.point, _so.point, _se.point );
				event.updateAfterEvent();
				
			}
		}

		private function onMouseUp( event: MouseEvent ): void
		{
			_mouse_down = false;
			_down_control = null;
			stage.removeEventListener( MouseEvent.MOUSE_MOVE, onMouseMove );
			stage.removeEventListener( MouseEvent.MOUSE_UP, onMouseUp );
		}

		private function onRemoved( event: Event ): void
		{
			stage.removeEventListener( MouseEvent.MOUSE_MOVE, onMouseMove );
			stage.removeEventListener( MouseEvent.MOUSE_UP, onMouseUp );
			target = null;
		}

		private function setNewTarget( value: DisplayObject ): void
		{
			_target = value;
			initializeTarget();
			
			if( value != null )
				dispatchEvent( new Event( "targetChange" ));
		}
	}
}